@@ -1,3 +1,4 @@
+import warnings
 import re
 
 def pytest_generate_tests(metafunc):
@@ -9,9 +10,9 @@ def pytest_generate_tests(metafunc):
         args = pymarks['scenario'].args
         scenario = args[0]
 
-        if len(args) > 1:
-            metafunc.config.warn(
-                0, 'WARNING - Only a subset of cases are selected: {}'.format(args)
+        if args[1] is not None:
+            warnings.warn(
+                'WARNING - Only a subset of cases are selected: {}'.format(args)
             )
             re_tag = re.compile(args[1])
             for name, _ in list(scenario.cases.items()):
